/* Copyright (c) 2000-2001 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

/*
The listener manager is responsible for all debugger listener operatations.
Don't confuse the confusing term "listener" here - the "listener" WRT this
file is what listens for remote TCP connections from external debuggees.

This listener manager also is responsible for the modeless dialog.  When you
start the listener, you pass a "timeout" period.  This is the number of seconds
before the dialog will display.  If you pass 0, it will display immediately.

As soon as a connection is established, the debugger callback is called, and the
window and listener are closed.  The user may also cancel the listener.

*/

function listener_event_handler(lum) {
    this.lum = lum;
}
listener_event_handler.prototype = {
	on_connect : function(language, host, port, cookie) {
		this.lum._onListenerConnect(language, host, port, cookie);
	},
	on_quit : function() {
		this.lum._onListenerQuit();
	},
	on_error : function(message) {
		this.lum._onListenerError(message);
	}
}


function listener_ui_manager(parent_window) {
    this.log = getLoggingMgr().getLogger("listener_ui_manager");
	// we may want multiple listeners later.
	this.listener = null;
	this.parent_window = parent_window;
	this.dialog = null; // xxx - probably should be replaced with code that finds the single "komodo:listener" window
	this.timeout_handle = null;
    this.globalPrefService = Components.classes["@activestate.com/koPrefService;1"].
        getService(Components.interfaces.koIPrefService);
    this.globalPrefs = this.globalPrefService.prefs;
}

listener_ui_manager.prototype = 
{
	startListening: function (ui_timeout, listenWindow) {
		if (!this.listener) {
			this.listener = Components.classes
			   ["@activestate.com/RemoteDebugConnectionListener;1"]
			   .getService(Components.interfaces.koIDebugConnectionListener);

			var handler = new listener_event_handler(this);
			this.listener.registerHandler(handler);
			this.error_message = null;
			
			// get the remote debug port from prefs, default if it doesn't exist
			var port = 9000;
			if (this.globalPrefs.hasPref("remoteDebuggerListenPort")) {
			    port = this.globalPrefs.getLongPref("remoteDebuggerListenPort");
			}
			var doPerlListen = false;
			if (this.globalPrefs.hasPref("perlListener")) {
			    doPerlListen = this.globalPrefs.getBooleanPref("perlListener");
			}
			var perlPort = 0;
			if (doPerlListen) {
			    perlPort = 9010;
			    if (this.globalPrefs.hasPref("remotePerlListenPort")) {
				perlPort = this.globalPrefs.getLongPref("remotePerlListenPort");
			    }
			}
			
			this.listener.listenNewThread(port,perlPort);
		    if (listenWindow) {
			  if (!ui_timeout) {
				this.showListenerDialog();
				this.timeout_handle = null;
			  } else if (listenWindow) {
				if (this.timeout_handle) {
					this.timeout_handle.stop();
					this.timeout_handle.free();
				}
				this.timeout_handle =
                    new objectTimer(this,this.showListenerDialog,[]);
                this.timeout_handle.startTimeout(ui_timeout);
			  }
		    }
		}

	},

	stopListening: function() {
		if (this.timeout_handle != null) {
		    this.timeout_handle.stopTimeout();
		    this.timeout_handle.free();
			this.timeout_handle = null;
		}
		if (this.listener != null) {
			try{
				this.listener.cancelListen();
			}catch( e ){
				this.log.error("Failed to cancel the listener: " + e);
			}
			this.listener = null;
		}
		if (this.error_message) {
			// being shutdown after an error - display the dialog
			// so the user can see the problem.
			if (this.dialog == null)
				this.showListenerDialog();
		} else {
			// Everything OK (either explicit cancel or connection) - close dialog.
			if (this.dialog)
				this.dialog.close();
		}
	},

	isListening: function() {
	    return this.listener != null;
	},
	
	_onListenerConnect: function(language, host, port, cookie) {
		this.listener = null;
		this.stopListening();
		this.parent_window.debugger_on_connection(language, host, port, cookie);
	},

	_onListenerError: function(message) {
		this.error_message = message;
		this.listener = null;
	},

	_onListenerQuit: function() {
		this.listener = null;
		this.stopListening();
	},

	showListenerDialog: function() {
		if (this.timeout_handle != null) {
		    this.timeout_handle.stopTimeout();
			this.timeout_handle = null;
		}
		 this.dialog = openWindowUniqueInstance(
			 "chrome://jaguar/content/debugger/listener.xul", 
			 "komodo:listener",
			 "chrome,dependent=1",
			  this);
	}

}
