// koLogging.js

// JS interface to the logging system.

/* This should be available in all JS code in Komodo.

Usage:

    log = gLoggingMgr.getLogger(<logger_name>);
    log.debug(<message>);
    log.info(<message>);
    etc.

To turn on or off a logger, use:

    log.setLevel(LOG_DEBUG)

Loggers that we are using:

    "ko.performance": Where major bits of performance information go.
    "ko.performance.startup": Perf logging w.r.t. startup perf.
    
    "ko.process": information re: the process module
*/

var _gLoggingMgr = null;

var LOG_NOTSET = 0;
var LOG_DEBUG = 10;
var LOG_INFO = 20;
var LOG_WARN = 30;
var LOG_ERROR = 40;
var LOG_CRITICAL = 50;

//XXX Whoa. Isn't this evil? This means every JS namespace has 'log'
//    set to the root logger -- a subtle fallback for sloppy code that
//    presumes a 'log' variable. --TM
var log;  // Default logger

// Logger wrapper objects get notified of the log levels associated with
// particular loggers.  They only make the calls to the logging system if
// those logging calls will actually be processed.

// See the 'set_logger_level' notification information

function loggerWrapper(logger, logger_name) {
    this.level = 0;
    this._logger = logger;
    this._logger_name = logger_name;
}

loggerWrapper.prototype.constructor = loggerWrapper;

loggerWrapper.prototype.setLevel = function(level) {
    this._logger.setLevel(level);
}

loggerWrapper.prototype.getEffectiveLevel = function() {
    return this._logger.getEffectiveLevel();
}

loggerWrapper.prototype.debug= function(message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_DEBUG) {
            this._logger.debug(message);
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.debug: "+ex+"\n");
    }
}

loggerWrapper.prototype.info = function(message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_INFO) {
            this._logger.info(message);
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.info: "+ex+"\n");
    }
}

loggerWrapper.prototype.warn = function(message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_WARN) {
            this._logger.warn(message);
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.warn: "+ex+"\n");
    }
}

loggerWrapper.prototype.error = function(message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_ERROR) {
            // I would prefer to have this be a separate log.exception(). --TM
            dump("Traceback from ERROR in '" +
                 this._logger_name + "' logger:\n    " +
                 gTimeline.getStack().replace('\n', '\n    ', 'g').slice(0, -4));
            this._logger.error(message);
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.error: "+ex+"\n");
    }
}

loggerWrapper.prototype.critical = function(message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_CRITICAL) {
            this._logger.critical(message);
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.critical: "+ex+"\n");
    }
}

loggerWrapper.prototype.exception = function(e, message) {
    try {
        if (this.level == 0) {
            this.level = this._logger.getEffectiveLevel();
        }
        if (this.level <= LOG_ERROR) {
            var objDump = gTimeline.dumpObjectTree(e,1);
            if (typeof(message)=='undefined' || !message)
                message='';
            this.error(message+'\n-- EXCEPTION START --\n'+objDump+'-- EXCEPTION END --\n');
        }
    } catch(ex) {
        dump("*** Error in loggerWrapper.critical: "+ex+"\n");
    }
}

function loggingMgr() {
    this.loggerMap = {}
    this.loggingSvc = Components.classes["@activestate.com/koLoggingService;1"].
                    getService(Components.interfaces.koILoggingService);
}

loggingMgr.prototype.constructor = loggingMgr;

loggingMgr.prototype.getLogger = function(logger_name) {
    if (logger_name in this.loggerMap) {
        return this.loggerMap[logger_name];
    }
    var logger = this.loggingSvc.getLogger(logger_name);
    this.loggerMap[logger_name] = new loggerWrapper(logger, logger_name);
    return this.loggerMap[logger_name];
}

loggingMgr.prototype.observe = function(subject, topic, data) {
    var logger_name;
    switch (topic) {
        case 'set_logger_level':
            logger_name = data.slice(0, data.indexOf(':'));
            var otherlogger, i;
            // We need to invalidate the logging levels of "derived" loggers.
            for (otherlogger in this.loggerMap) {
                if (logger_name == '' || logger_name == otherlogger || otherlogger.indexOf(logger_name+'.') == 0) {
                    this.loggerMap[otherlogger].level = 0;
                }
            }
            var log_level;
            log_level = Number(data.slice(data.indexOf(':')+1));
            this.getLogger(logger_name).level = log_level; // bypass the 'setLevel call'
            break;
    }
}

// Use this function to always get the logging manager
// (which may not be a global in your namespace)
function getLoggingMgr() {
    // Factory function which returns the "superglobal" loggingMgr instance
    var namespace = window;
    function haveParent(namespace) {
        return ('parent' in namespace && namespace.parent
                && namespace.parent != namespace);
    }
    function haveOpener(namespace) {
        return ('opener' in namespace && namespace.opener);
    }
    while (namespace._gLoggingMgr == null
           && (haveParent(namespace) || haveOpener(namespace))) {
        if (haveParent(namespace)) {
            namespace = namespace.parent;
        } else {
            namespace = namespace.opener;
        }
    }

    if (namespace._gLoggingMgr == null) {
        // No logging manager has yet been created. Create it.
        namespace._gLoggingMgr = new loggingMgr();
        var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                                getService(Components.interfaces.nsIObserverService);
        observerSvc.addObserver(namespace._gLoggingMgr, 'set_logger_level',false);
    }
    return namespace._gLoggingMgr;
}

try {
    log = getLoggingMgr().getLogger('');
} catch (e) {
    alert(e);
}


