import fnmatch
from types import *
import timeline

from xpcom import components, nsError, ServerException
import xpcom.server
from xpcom.server.enumerator import SimpleEnumerator
import re
import logging
log = logging.getLogger('koLanguage')

class koLanguageItem:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koLanguageItem;1"
    _reg_clsid_ = "{33532bc1-302a-4b2d-ad14-13c5eadf4d93}"
    
    def __init__(self, language):
        self.language = language
        self.name = language
        
    def get_available_types(self):
        return components.interfaces.koIHierarchyItem.ITEM_STRING

    def get_item_string(self):
        return self.language

    def get_container(self):
        return 0

class koLanguageContainer:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koLanguageContainer;1"
    _reg_clsid_ = "{878ed885-6274-4c07-9668-a9a01a0ae09c}"
    
    def __init__(self, label, languages):
        self.name = label
        self.languages = languages
    
    def getChildren(self):
        return self.languages

    def get_available_types(self):
        return 0
        return components.interfaces.koIHierarchyItem.ITEM_STRING

    def get_item_string(self):
        return 0
        return self.label

    def get_container(self):
        return 1
    
# The LanguageRegistryService keeps track of which languages/services
# are available.  It is used by packages of services to inform the
# system of their presence, and is used internally by the other
# language services classes to get information on the services
# available.

class koLanguageRegistryService:
    _com_interfaces_ = [components.interfaces.koILanguageRegistryService,
                        components.interfaces.nsIObserver]
    _reg_contractid_ = "@activestate.com/koLanguageRegistryService;1"
    _reg_clsid_ = "{4E76795E-CC92-47c6-8801-C9ACFC1B02E3}"

    primary = ['HTML', 'JavaScript', 'PHP', 'Perl', 'Python', 'Tcl', 'Text', 'XML', 'XSLT']
    internal = ['Rx']
    
    # XXX this belongs in language services
    shebangPatterns = [
        ('Perl', re.compile(ur'^#!.*perl.*$', re.IGNORECASE)),
        ('PHP', re.compile(ur'^#!.*php.*$', re.IGNORECASE)),
        ('Python', re.compile(ur'^#!.*python.*$', re.IGNORECASE)),
        ('Ruby', re.compile(ur'^#!.*ruby.*$', re.IGNORECASE)),
        ('Tcl', re.compile(ur'^#!.*expect.*$', re.IGNORECASE)),    
        ('Tcl', re.compile(ur'^#!.*tcl.*$', re.IGNORECASE)),    
        ('Tcl', re.compile(ur'^#!.*wish.*$', re.IGNORECASE)),    
    ]
    
    re_firstline = re.compile(ur'(.*?)(?:\r|\n|$)')
    
    def __init__(self):
        timeline.enter('koLanguageRegistryService.__init__')
        self.__languages = {}
        self.__accesskeys = {}
        self.__filters = None
        self.__filtersByLanguage = None
        self.__associations = None
        self.__absolutefilters = {}
        self.__initPrefs()
        self._resetFileAssociations()

        # XXX get from prefs?
        self.defaultLanguage = "Text"
        
        self.registerLanguage("C++",'c');
        self.registerLanguage("CSS",'');
        self.registerLanguage("HTML",'m');
        self.registerLanguage("JavaScript",'a');
        self.registerLanguage("Perl",'p');  
        self.registerLanguage("Python",'y');
        self.registerLanguage("Tcl",'t');
        self.registerLanguage("Text",'e');
        self.registerLanguage("XML",'x');
        self.registerLanguage("XSLT",'s');
        self.registerLanguage("Pascal",'');
        self.registerLanguage("PHP",'h');
        self.registerLanguage("Java",'j');
        self.registerLanguage("C#",'#');
        self.registerLanguage("Lua",'u');
        self.registerLanguage("Makefile",'k');
        self.registerLanguage("IDL",'i');
        self.registerLanguage("Diff",'d');
        self.registerLanguage("Batch",'b');
        self.registerLanguage("LaTeX",'l');
        self.registerLanguage("Lisp",'');
        self.registerLanguage("Ada",'D');
        self.registerLanguage("Ruby",'');
        self.registerLanguage("VisualBasic",'v');
        self.registerLanguage("SQL",'q');
        self.registerLanguage("Eiffel",'f');
        self.registerLanguage("Baan",'');
        self.registerLanguage("nnCrontab",'');
        self.registerLanguage("Matlab",'');
        self.registerLanguage("Bullant",'');
        self.registerLanguage("Apache",'');
        self.registerLanguage("Fortran 77",'');
        self.registerLanguage("Fortran",'');
        self.registerLanguage("Errors",'');
        self.registerLanguage("Rx",'');
        timeline.leave('koLanguageRegistryService.__init__')
    
    def observe(self, aSubject, aTopic, someData):
        if someData == "fileAssociations":
            self.__filters = None
            self.__filtersByLanguage = None
            self.__associations = []
            self._resetFileAssociations()
            
    def _resetFileAssociations(self):
        timeline.enter('koLanguageRegistryService._resetFileAssociations')
        associations = self.__prefs.getPref("fileAssociations") \
                       .QueryInterface(components.interfaces.koIOrderedPreference)
        self.__filters = []
        self.__filtersByLanguage = {}
        self.__associations = []
        self.__absolutefilters = {}
        for i in range(0, associations.length):
            association = associations.getPref(i) \
                          .QueryInterface(components.interfaces.koIOrderedPreference)
            filterMatch = association.getStringPref(0)
            filterLanguage = association.getStringPref(1)

            # set filters
            self.__filters.append(filterMatch)

            # set filters by language
            if not self.__filtersByLanguage.has_key(filterLanguage):
                self.__filtersByLanguage[filterLanguage] = []
            self.__filtersByLanguage[filterLanguage].append(filterMatch)

            # get the file extensions only
            if filterMatch[:2] == "*.":
                self.__absolutefilters[filterMatch[2:]] = filterLanguage
            self.__absolutefilters[filterMatch] = filterLanguage
            
            # set associations
            self.__associations.append((filterMatch,filterLanguage))
        timeline.leave('koLanguageRegistryService._resetFileAssociations')

    def getLanguageHierarchy(self):
        # We could determine primary languages programmatically but seeing which
        # features each language offers, but that would involve creating an
        # XPCOM object for each language.

        secondary = [i for i in self.__languages.keys()
                        if i not in self.primary and
                           i not in self.internal]
        secondary.sort()

        otherContainer = koLanguageContainer('Other', [koLanguageItem(i) for i in secondary])
        return koLanguageContainer('', [koLanguageItem(i) for i in self.primary] + [otherContainer])
                
    def getLanguages(self):

        languages = self.__languages.keys()
        languages.sort()
        languages = [language for language in languages if language not in self.internal]
        return languages
    
    def getFilters(self):
        if self.__filters is None:
            self._resetFileAssociations()

        return self.__filters

    def getFiltersForLanguage(self, language):
        if self.__filtersByLanguage is None:
            self._resetFileAssociations()

        if self.__filtersByLanguage.has_key(language):
            filters = self.__filtersByLanguage[language]
        else:
            filters = []

        return filters

    def registerLanguage(self, language, accesskey):
        assert not self.__languages.has_key(language), \
               "Language '%s' already registered" % (language)
        
        self.__languages[language] = None
        self.__accesskeys[language] = accesskey

    def getAccessKeys(self):
        languages = self.__languages.keys()
        languages.sort()
        return [self.__accesskeys[language] for language in languages]
    
    def getLanguage(self, language):
        # return a koILanguage for language.  Create it if it does not
        # exist yet.
        if not language: language=self.defaultLanguage
        
        if language not in self.__languages:
            log.warn("Asked for unknown language: %r", language)
            return None

        if self.__languages[language] is None:
            contractid = "@activestate.com/koLanguage?language=%s;1" \
                         % (language)
            self.__languages[language] = components.classes[contractid] \
                    .createInstance(components.interfaces.koILanguage)

        return self.__languages[language]

    def suggestLanguageForFile(self, filename):
        if self.__associations is None:
            self._resetFileAssociations()

        language = ''  # default

        # get the file extension
        parts = filename.split(".")

        if len(parts) > 0:
            ext = parts[len(parts)-1];
            if self.__absolutefilters.has_key(ext):
                return self.__absolutefilters[ext]

        if self.__absolutefilters.has_key(filename):
            return self.__absolutefilters[filename]
        
        # check for file pattern matches        
        for (pat, lang) in self.__associations:
            if fnmatch.fnmatch(filename, pat):
                language = lang
                break
            
        return language

    def guessLanguageFromContents(self, buffer, default):
        if not default:
            default = self.defaultLanguage
        if buffer.startswith(u'<?xml'): # Found an XML declaration
            return ['XML', self.defaultLanguage, 'XSLT']
        elif buffer.startswith(u'#!'): # Found shebang line
            buffer = self.re_firstline.match(buffer).groups(0)[0]
            possibleLanguages = []
            for language, pattern in self.shebangPatterns:
                if pattern.match(buffer):
                    possibleLanguages.append(language)                
            if len(possibleLanguages) == 0:
                return []
            elif len(possibleLanguages) == 1:
                possibleLanguages.append(default)
                return possibleLanguages
            else:
                possibleLanguages.insert(0,None)
                possibleLanguages.append(default)
                return possibleLanguages
        else:
            return []
        
    def __initPrefs(self):
        self.__prefs = components \
                       .classes["@activestate.com/koPrefService;1"] \
                       .getService(components.interfaces.koIPrefService)\
                       .prefs
        # Observers will be QI'd for a weak-reference, so we must keep the
        # observer alive ourself, and must keep the COM object alive,
        # _not_ just the Python instance!!!
        # XXX - this is a BUG in the weak-reference support.
        # It should NOT be necessary to do this, as the COM object is
        # kept alive by the service manager.  I suspect that this bug
        # happens due to the weak-reference being made during
        # __init__.  FIXME!
        self._observer = xpcom.server.WrapObject(self,
                                      components.interfaces.nsIObserver)
        self.__prefs.addObserver(self._observer)



class koSourceFile:
    _com_interfaces_ = [components.interfaces.koISourceFile]
    _reg_contractid_ = "@activestate.com/koSourceFile;1"
    _reg_clsid_ = "{D579D901-7DBF-4fa6-AE7A-1A85DAC9CA83}" 

    _registry = None
    def __init__(self):
        if koSourceFile._registry is None:
            koSourceFile._registry = components.classes["@activestate.com/koLanguageRegistryService;1"].getService(components.interfaces.koILanguageRegistryService) 
        
    def init(self, scintilla):
        self._scin = scintilla
        self.suggestLanguage("Text")

    def getLanguageBlocks(self):
        #return SimpleEnumerator(self.__languageBlocks)
        return self._languageBlocks

    def getLanguageBlockAtPosition(self, position):
        # for now, we only allow one language block per file
        assert 0 <= position <= self._scin.length, "position out of range"
        assert len(self._languageBlocks) == 1, "internal inconsistency"
        
        return self._languageBlocks[0]

    def suggestLanguage(self, language):
        self._languageBlocks = [
            self._createLanguageBlock(language, 0, self._scin.length)
            ]
        
    def suggestFilename(self, basename):
        # make the association and create a language block appropriately
        language = koSourceFile._registry.suggestLanguageForFile(basename)
        self.suggestLanguage(language or 'Text')

    def _createLanguageBlock(self, language, offset, length):
        return koLanguageBlock(self._scin, language, offset, length, koSourceFile._registry)


class koLanguageBlock:
    _com_interfaces_ = [components.interfaces.koILanguageBlock]

    def __init__(self, scintilla, language, offset, length, registry):
        # for now, we only allow one language block per file
        assert offset == 0
        assert length == scintilla.length
        
        self.__scin = scintilla
        self.offset = offset
        self.length = length
        self.__language = language
        self.__registry = registry


    def get_language(self):
        return self.__registry.getLanguage(self.__language)
    
    def get_text(self):
        assert self.offset+self.length <= self.__scin.length
        
        return self.__scin.text[self.offset:self.offset+self.length]


    def getRange(self, start, end):
        assert self.offset+start <= self.__scin.length
        assert self.offset+end <= self.__scin.length
        assert start <= end
        
        return self.get_text()[start:end]


