# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components
sci_constants = components.interfaces.ISciMoz
from xpcom.server import WrapObject, UnwrapObject
import logging
import re
log = logging.getLogger('koLanguageCommandHandler')
indentlog = logging.getLogger('koLanguageCommandHandler.indenting')

import string
import timeline

"""
The generic command handler is appropriate for all languages.

  It handles things like:
    block selection
    brace matching and jumps to matching brace
    transpose
    ctrl-space
    indentation control
    
"""

class GenericCommandHandler:
    _com_interfaces_ = [components.interfaces.koIViewController,
                        components.interfaces.nsIController,
                        components.interfaces.nsIObserver]
    _reg_desc_ = "Command Handler for All files"
    _reg_contractid_ = "@activestate.com/koGenericCommandHandler;1"
    _reg_clsid_ = "{B383592C-5343-4AA7-9419-04D1B34EC906}"
    
    def __init__(self):
        log.info("in __init__ for GenericCommandHandler")
        self._completeWordState = None
        self._view = None
        self.sysUtils = components.classes["@activestate.com/koSysUtils;1"].\
            getService(components.interfaces.koISysUtils)

    def __del__(self):
        log.info("in __del__ for GenericCommandHandler")
        
    def set_view(self, view):
        if view:
            self._view = view.QueryInterface(components.interfaces.koIScintillaView)
            self._prefs = view.prefs
            self._scimoz = self._view.scimoz
        else:
            self._view = None
            self._prefs = None
            self._scimoz = None

    def get_view(self):
        return self._view

    def _is_cmd_fontFixed_enabled(self):
        return 1

    def _do_cmd_fontFixed(self):
        self._view.alternateFaceType = not self._view.alternateFaceType
        self._view.prefs.setBooleanPref('editUseAlternateFaceType',
                                        not self._view.prefs.getBooleanPref('editUseAlternateFaceType'))

    def _is_cmd_viewWhitespace_enabled(self):
        return self._scimoz.viewWS

    def _do_cmd_viewWhitespace(self):
        self._view.scimoz.viewWS = not self._view.scimoz.viewWS

    def _is_cmd_viewLineNumbers_enabled(self):
        return self._scimoz.getMarginWidthN(0) > 0

    def _updateLineNumberMargin(self):
        numLinesToAccountFor = max(1000,self._scimoz.lineCount*2)
        padding = 5
        textWidth = self._scimoz.textWidth(0, str(numLinesToAccountFor))
        self._scimoz.setMarginWidthN(0, textWidth + padding)

    def _do_cmd_viewLineNumbers(self):
        shown = self._scimoz.getMarginWidthN(0) > 0
        shown = not shown
        if shown:
            self._updateLineNumberMargin()
        else:
            self._scimoz.setMarginWidthN(0, 0)

    def _is_cmd_viewIndentationGuides_enabled(self):
        return self._scimoz.indentationGuides

    def _do_cmd_viewIndentationGuides(self):
        self._scimoz.indentationGuides = not self._scimoz.indentationGuides

    def _is_cmd_viewEOL_enabled(self):
        return self._scimoz.viewEOL

    def _do_cmd_viewEOL(self):
        self._scimoz.viewEOL = not self._scimoz.viewEOL

    def _is_cmd_wordWrap_enabled(self):
        return self._scimoz.wrapMode

    def _do_cmd_wordWrap(self):
        if self._scimoz.wrapMode == self._scimoz.SC_WRAP_NONE:
            self._scimoz.wrapMode = self._scimoz.SC_WRAP_WORD
            self._scimoz.layoutCache = self._scimoz.SC_CACHE_PAGE
        else:
            self._scimoz.wrapMode = self._scimoz.SC_WRAP_NONE
            self._scimoz.layoutCache = self._scimoz.SC_CACHE_NONE

    def isCommandEnabled( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        meth = getattr(self, "_is_%s_enabled" % (str(command_name),), None)
        if meth is None:
            rc = 1
        else:
            rc = meth()
        log.debug("koLanguageCommandHandler - isCommandEnabled '%s' -> %d" % (command_name, rc))
        return rc

    def supportsCommand( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        rc = hasattr(self, "_do_" + command_name)
        log.debug("koLanguageCommandHandler - SupportsCommand '%s' -> %d" % (command_name, rc))
        return rc

    def doCommand( self, command_name ):
        # Result: void - None
        # In: param0: wstring
        meth = getattr(self, "_do_" + command_name)
        try:
            # Some CommandHandler's do not have a wrapper, e.g. TerminalCommandHandler
            #if self.wrapper is not None:
            #    self.wrapper.beginBatchOperation()
            #try:
                meth()
            #finally:
                #if self.wrapper is not None:
                #    self.wrapper.endBatchOperation()
        except KeyboardInterrupt: # User cancelled a dialog.
            pass

    def _getFoldLevel(self, line):
        """Return a fold level for the current line that guarantees that:
            - if at the top level, the fold level is zero
            - the deeper you are the higher the fold level
        It does *not* guarantee that:
            - a depth of three means then level is three
        """ 
        # if the line being asked about has never been lexed, then lex it.
        _startOfLine = self._scimoz.positionFromLine(line)
        _style = self._scimoz.getStyleAt(_startOfLine)
        if _style == 0:
            # NOTE: assumption that style = 0 means that it hasn't been lexed.
            #_startOfNextLine = self._scimoz.positionFromLine(line+20)
            self._scimoz.colourise(0, -1)

        topFoldLevel = self._scimoz.getFoldLevel(0)
        topFoldLevel &= self._scimoz.SC_FOLDLEVELNUMBERMASK
        foldLevel = self._scimoz.getFoldLevel(line)
        foldLevel &= self._scimoz.SC_FOLDLEVELNUMBERMASK
        foldLevel -= topFoldLevel
        return foldLevel

    def _do_cmd_newlineExtra(self):
        self._do_cmd_newline(None, 1)

    def _do_cmd_newlineBare(self):
        self._do_cmd_newline('none', 0)

    def _do_cmd_newlineSame(self):
        self._do_cmd_newline('plain', 0)

    def _do_cmd_newline(self, indentStyle=None, continueComments=0):
        if indentStyle is None:
            indentStyle = self._prefs.getStringPref('editAutoIndentStyle')
        sm = self._scimoz
        sm.beginUndoAction()
        try:
            try:
                if indentStyle == 'none':
                    indent = ''
                else:
                    indent = self._view.languageObj.computeIndent(self._scimoz,
                                                                  indentStyle,
                                                                  continueComments)
                # Optionally clean up whitespace characters to the left of cursor if there
                # is something else there (leave whitespace-only lines alone, as otherwise
                # typing { <cr><cr> } up-arrow is frustrating.
                lineNo = sm.lineFromPosition(sm.currentPos)
                lineStart = sm.positionFromLine(lineNo)
                stuffToLeft = sm.getTextRange(lineStart, sm.currentPos)
                if stuffToLeft.strip():
                    while (sm.currentPos > lineStart and
                           sm.getWCharAt(sm.positionBefore(sm.currentPos)) in ' \t'):
                        sm.deleteBack()
                    # OPTIONALLY delete whitespace to the right of the 
                    # current cursor position
                    if 1: # XXX make a pref, give a UI.
                        lineEnd = sm.getLineEndPosition(lineNo)
                        textToRight = sm.getTextRange(sm.currentPos, lineEnd)
                        while sm.getWCharAt(sm.currentPos) in ' \t':
                            sm.clear()
                
                sm.newLine()
                if indent:
                    sm.addText(self.sysUtils.byteLength(indent), indent) # consider using replaceTarget??
                sm.chooseCaretX()
            except Exception, e:
                log.exception("Error in cmd_newline handling")
        finally:
            sm.endUndoAction()
        

    def _getIndentWidthForLine(self, lineNo, upto=None):
        indent = self._getIndentForLine(lineNo, upto)
        return len(indent.expandtabs(self._scimoz.tabWidth))

    def _getIndentForLine(self, lineNo, upto=None):
        sm = self._scimoz
        lineStart = sm.positionFromLine(lineNo)
        if upto:
            lineEnd = min(upto, sm.getLineEndPosition(lineNo))
        else:
            lineEnd = sm.getLineEndPosition(lineNo)
        line = sm.getTextRange(lineStart, lineEnd)
        indentLength = len(line)-len(line.lstrip())
        return line[:indentLength]

    def _getIndentWidthFromIndent(self, indentstring):
        """ Given an indentation string composed of tabs and spaces,
        figure out how long it is (get the tabWidth from scimoz) """
        expandedstring = indentstring.expandtabs(self._scimoz.tabWidth)
        return len(expandedstring)

    def _makeIndentFromWidth(self, width):
        """ returns a string consisting of however many tabs and spaces
        are needed to make a width of `width`, using the useTabs pref
        and the scimoz tabWidth
        """
        if self._scimoz.useTabs:
            numtabs, numspaces = divmod(width, self._scimoz.tabWidth)
            return '\t'*numtabs + ' '*numspaces
        else:
            return ' '*width

    def _do_cmd_blockSelect(self):
        """Use the lexer's fold markers to increase the selection.
        If the current selection *is* a complete "fold block" then increase
        to the next fold block. If the current selection is *not* a complete
        fold block then select the fold block of the shallowest end of the
        selection.

        The logic is complicated somewhat because, the line preceeding a
        block at a certain level should be included in the block selection.
        For example, in the following code, if "Ctrl-B" is hit with the
        cursor on the "print 1" line, the "for element in mylist:" line
        should be included in the selection.

        LEVEL   CODE
        0       def foo(a):
        1           mylist = ["hello", 42, "there", 3.14159]
        1           while 1:
        2               for element in mylist:
        3                   print 1
        3                   if type(element) == type(""):
        4                       print "element %s is a string" % element
        1           print
        """
        #log.debug("----------------- BLOCK SELECT ---------------------")
        # Ensure cursor is at the start of the current selection (so we
        # don't have to deal with the reverse condition in the calculations
        # below.
        if self._scimoz.anchor < self._scimoz.currentPos:
            self._scimoz.anchor, self._scimoz.currentPos =\
                self._scimoz.currentPos, self._scimoz.anchor

        startLine = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        endLine = self._scimoz.lineFromPosition(self._scimoz.anchor)
        lastLine = self._scimoz.lineFromPosition(self._scimoz.textLength-1)
        
        # if the lexer doesn't support folding, select all and get out of here fast.
        lexer = self._view.languageObj.getLanguageService(
            components.interfaces.koILexerLanguageService)
        if not lexer.supportsFolding:
            self._scimoz.selectAll()
            return

        # If the next line after the startLine is "deeper" then move the
        # startLine to there. See the doc string for why this is done.
        if (startLine != lastLine and
            self._getFoldLevel(startLine+1) > self._getFoldLevel(startLine)):
            startLine += 1

        # If there is a selection and selectionEnd is at the start of a line
        # then move up the end of the previous line because this simplifies
        # the subsequent logic.
        if (self._scimoz.selText != ""
            and self._scimoz.getColumn(self._scimoz.anchor) == 0):
            endLine -= 1
            self._scimoz.anchor = self._scimoz.getLineEndPosition(endLine)
        #log.debug("BLOCKSELECT: startLine=%d, endLine=%d, lastLine=%d",
        #          startLine, endLine, lastLine)

        # Determine the minimum fold level in the current selection.
        minFoldLevel = self._getFoldLevel(startLine)
        for line in range(startLine, endLine):
            #log.debug("BLOCKSELECT: foldLevel for line %d is: %x",
            #          line, self._getFoldLevel(line))
            minFoldLevel = min(minFoldLevel, self._getFoldLevel(line))
        #log.debug("BLOCKSELECT: minFoldLevel is %x" % minFoldLevel)

        # Determine if the current selection is a complete fold block.
        # - selection start must be at the start of a line
        # - selection end must be at the end of a line
        # - start and end fold level must both be at the bound of the
        #   minFoldLevel
        if (self._scimoz.getColumn(self._scimoz.currentPos) == 0
            and self._scimoz.anchor == self._scimoz.getLineEndPosition(endLine)
            and (startLine == 0
                 or (self._getFoldLevel(startLine-1) < minFoldLevel
                     and self._getFoldLevel(startLine)
                         > self._getFoldLevel(startLine-1))
                )
            and (endLine == lastLine
                 or (self._getFoldLevel(endLine+1) < minFoldLevel
                     and self._getFoldLevel(endLine)
                         > self._getFoldLevel(endLine+1))
                )
           ):
            #log.debug("BLOCKSELECT: *is* a complete fold block")
            if startLine == 0:
                targetFoldLevel = 0
            else:
                targetFoldLevel = self._getFoldLevel(startLine-1)
        else:
            #log.debug("BLOCKSELECT: condition: start of selection on column 0? %s",
            #          self._scimoz.getColumn(self._scimoz.currentPos) == 0)
            #log.debug("BLOCKSELECT: condition: end of selection at eol? %s",
            #          self._scimoz.anchor == self._scimoz.getLineEndPosition(endLine))
            #log.debug("BLOCKSELECT: condition: startLine at minFoldLevel boundary? %s",
            #          (startLine == 0 or
            #           (self._getFoldLevel(startLine-1) < minFoldLevel
            #            and self._getFoldLevel(startLine)
            #                > self._getFoldLevel(startLine-1))))
            #log.debug("BLOCKSELECT: condition: endLine at minFoldLevel boundary? %s",
            #          (endLine == lastLine or
            #           (self._getFoldLevel(endLine+1) < minFoldLevel
            #            and self._getFoldLevel(endLine)
            #                > self._getFoldLevel(endLine+1))))
            #log.debug("BLOCKSELECT: is *not* a complete fold block")
            targetFoldLevel = minFoldLevel
        #log.debug("BLOCKSELECT: targetFoldLevel is %x" % targetFoldLevel)

        # increase the selection to the target fold level
        if targetFoldLevel == 0:
            #log.debug("BLOCKSELECT: select all")
            self._scimoz.anchor = self._scimoz.textLength
            self._scimoz.currentPos = 0
            self._scimoz.ensureVisibleEnforcePolicy(0)
        else:
            # expand upwards
            while startLine > 0:  #XXX are line's in SciMoz 0- or 1-based???
                #log.debug("BLOCKSELECT: level(%x) up one line(%d) > target level(%x)?",
                #          self._getFoldLevel(startLine-1), startLine-1,
                #          targetFoldLevel)
                if self._getFoldLevel(startLine-1) >= targetFoldLevel:
                    startLine -= 1
                else:
                    break 
            # select the leading line in the block as well
            if startLine != 0:
                startLine -= 1

            # expand downwards
            while endLine < lastLine:
                #log.debug("BLOCKSELECT: level(%x) down one line(%d) > target level(%x)?",
                #          self._getFoldLevel(endLine+1), endLine+1,
                #          targetFoldLevel)
                if self._getFoldLevel(endLine+1) >= targetFoldLevel:
                    endLine += 1
                else:
                    break 

            # make the selection
            #log.debug("BLOCKSELECT: startLine=%d, endLine=%d",
            #          startLine, endLine)
            if endLine == lastLine:
                self._scimoz.anchor = self._scimoz.textLength
            else:
                self._scimoz.anchor = self._scimoz.positionFromLine(endLine+1)
            self._scimoz.currentPos = self._scimoz.positionFromLine(startLine)
            self._scimoz.ensureVisibleEnforcePolicy(startLine)

    def _is_cmd_comment_enabled(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_comment(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        commenter.comment(self._scimoz)

    def _is_cmd_uncomment_enabled(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_uncomment(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        commenter.uncomment(self._scimoz)

    def _do_cmd_convertUpperCase(self):
        selStart = self._scimoz.selectionStart
        selEnd = self._scimoz.selectionEnd
        self._scimoz.replaceSel(self._scimoz.selText.upper())
        self._scimoz.selectionStart = selStart
        self._scimoz.selectionEnd = selEnd

    def _do_cmd_convertLowerCase(self):
        selStart = self._scimoz.selectionStart
        selEnd = self._scimoz.selectionEnd
        self._scimoz.replaceSel(self._scimoz.selText.lower())
        self._scimoz.selectionStart = selStart
        self._scimoz.selectionEnd = selEnd

    def _is_cmd_selectToMatchingBrace_enabled(self):
        matchingBrace = self._scimoz.braceMatch(self._scimoz.positionBefore(self._scimoz.currentPos))
        log.info("matchingBrace @ %d : %d" % (self._scimoz.currentPos, matchingBrace))
        if matchingBrace != -1: return 1 # we're there
        # also check to the right
        if self._scimoz.currentPos < self._scimoz.textLength:
            matchingBrace = self._scimoz.braceMatch(self._scimoz.currentPos)
            log.info("matchingBrace @ %d : %d" % (self._scimoz.currentPos, matchingBrace))
            return matchingBrace != -1
        else:
            log.info("matchingBrace: returning 0") 
            return 0

    _is_cmd_jumpToMatchingBrace_enabled = _is_cmd_selectToMatchingBrace_enabled
    _is_cmd_braceMatch_enabled = _is_cmd_selectToMatchingBrace_enabled

    def _do_cmd_braceMatch(self):
        braceAtCaret, braceOpposite, isInside = self._findMatchingBracePosition(self._scimoz.currentPos)
        
        if braceAtCaret != -1 and braceOpposite == -1:
            self._scimoz.braceBadLight(braceAtCaret)
            self._scimoz.highlightGuide = 0
        else:
            chBrace = self._scimoz.getWCharAt(braceAtCaret)
            self._scimoz.braceHighlight(braceAtCaret, braceOpposite)
            columnAtCaret = self._scimoz.getColumn(braceAtCaret)
            if chBrace == ':':
                lineStart = self._scimoz.lineFromPosition(braceAtCaret)
                indentPos = self._scimoz.getLineIndentPosition(lineStart)
                indentPosNext = self._scimoz.getLineIndentPosition(lineStart+1)
                columnAtCaret = self._scimoz.getColumn(indentPos)
                columnAtCaretNext = self._scimoz.getColumn(indentPosNext)
                indentSize = self._scimoz.indent
                if (columnAtCaretNext - indentSize > 1):
                    columnAtCaret = columnAtCaretNext - indentSize 
            columnOpposite = self._scimoz.getColumn(braceOpposite)
            self._scimoz.highlightGuide = min(columnAtCaret, columnOpposite)

    def _do_cmd_selectToMatchingBrace(self):
        self._goMatchingBrace(1)

    def _do_cmd_jumpToMatchingBrace(self):
        self._goMatchingBrace(0)

    def _goMatchingBrace(self, select):
        braceAtCaret, braceOpposite, isInside = self._findMatchingBracePosition(self._scimoz.currentPos)
        log.info("braceAtCaret: %d, braceOpposite: %d, isInside: %s" % (braceAtCaret, braceOpposite, isInside))
        # Convert the character positions into caret positions based on whether
        # the caret position was inside or outside the braces.
        if isInside:
            if braceOpposite > braceAtCaret:
                braceAtCaret += 1
            else:
                braceOpposite += 1
        else: # Outside
            if braceOpposite > braceAtCaret:
                braceOpposite += 1
            else:
                braceAtCaret += 1
        if braceOpposite >= 0:
            self._ensureRangeVisible(braceOpposite, braceOpposite)
            if select:
                self._scimoz.anchor = braceAtCaret
                self._scimoz.currentPos = braceOpposite
            else:
                self._scimoz.anchor = braceOpposite
                self._scimoz.currentPos = braceOpposite

    def _ensureRangeVisible(self, posStart, posEnd, enforcePolicy=1):
        lineStart = self._scimoz.lineFromPosition(min(posStart, posEnd))
        lineEnd = self._scimoz.lineFromPosition(max(posStart, posEnd))
        for line in range(lineStart, lineEnd+1):
            if enforcePolicy:
                self._scimoz.ensureVisibleEnforcePolicy(line)
            else:
                self._scimoz.ensureVisible(line)

    # Find the brace that matches the brace before or after the given caret
    # position. The character before has precedence if it is a brace.
    # Returns three values: whether the caret is inside the braces,
    # and the indexes of the current and the matching brace.
    def _findMatchingBracePosition(self, caretPos, sloppy=1):
        mask = self._view.languageObj.stylingBitsMask
        isInside = 0
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = '\0'
        styleBefore = 0
        textLength = self._scimoz.textLength
        if caretPos > textLength:
            caretPos = textLength
        if (caretPos > 0) :
            charBefore = self._scimoz.getWCharAt(self._scimoz.positionBefore(caretPos))
            styleBefore = self._scimoz.getStyleAt(self._scimoz.positionBefore(caretPos)) & mask
        # Priority goes to character before caret
        if charBefore and self._view.languageObj.getBraceIndentStyle(charBefore, styleBefore):
            braceAtCaret = caretPos - 1
        
        colonMode = 0
        if self._view.languageObj.name == 'Python' and ':' == charBefore:
            braceAtCaret = caretPos - 1
            colonMode = 1

        isAfter = 1
        if sloppy and (braceAtCaret < 0) and \
           0 <= caretPos < textLength: # XXX check last edge condition
            # No brace found so check other side
            charAfter = self._scimoz.getWCharAt(caretPos)
            styleAfter = self._scimoz.getStyleAt(caretPos) & mask
            
            if charAfter and (charAfter in "[](){}") and \
               self._view.languageObj.getBraceIndentStyle(charAfter, styleAfter):
                braceAtCaret = caretPos
                isAfter = 0

            if self._view.languageObj.name == 'Python' and ':' == charAfter:
                braceAtCaret = caretPos
                colonMode = 1
                
        if braceAtCaret >= 0:
            if colonMode:
                lineStart = self._scimoz.lineFromPosition(braceAtCaret)
                lineMaxSubord = self._scimoz.getLastChild(lineStart, -1)
                braceOpposite = self._scimoz.getLineEndPosition(lineMaxSubord)
            else:
                braceOpposite = self._scimoz.braceMatch(braceAtCaret)
                
        if braceOpposite > braceAtCaret:
            isInside = isAfter
        else:
            isInside = not isAfter
        return braceAtCaret, braceOpposite, isInside

    def _is_cmd_folding_enabled(self):
        return self._view.languageObj.foldable

    _is_cmd_foldExpand_enabled = _is_cmd_folding_enabled
    _is_cmd_foldExpandAll_enabled = _is_cmd_folding_enabled
    _is_cmd_foldCollapse_enabled = _is_cmd_folding_enabled
    _is_cmd_foldCollapseAll_enabled = _is_cmd_folding_enabled
    
    def _do_cmd_foldExpand(self):
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos);
        if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
            not self._scimoz.getFoldExpanded(lineno):
            self._scimoz.toggleFold(lineno)

    def _do_cmd_foldExpandAll(self):
        for lineno in range(0, self._scimoz.lineCount):
            if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
                not self._scimoz.getFoldExpanded(lineno):
                    self._scimoz.toggleFold(lineno)

    def _do_cmd_foldCollapse(self):
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        while not (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG):
            if lineno == 1 or not self._scimoz.getFoldExpanded(lineno):
                return
            lineno -= 1
        self._scimoz.toggleFold(lineno)
        self._scimoz.gotoLine(lineno)

    def _do_cmd_foldCollapseAll(self):
        self._scimoz.colourise(0,-1)
        for lineno in range(0, self._scimoz.lineCount):
            if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
                self._scimoz.getFoldExpanded(lineno):
                self._scimoz.toggleFold(lineno)


    def _asktabwidth(self):
        dialogproxy = components.classes['@activestate.com/asDialogProxy;1'].\
            getService(components.interfaces.asIDialogProxy)
        tabwidth = self._view.prefs.getLongPref('tabWidth')
        value = dialogproxy.prompt("Tab Width (between 0 and 16)", str(tabwidth), "OK", None)
        if value is not None:
            return int(value)
        else:
            return value
                                  

    def _do_cmd_tabify(self):
        selection = self._scimoz.selText
        tabwidth = self._asktabwidth()
        if tabwidth is None: return
        lines = selection.splitlines(1)
        for pos in range(len(lines)):
            line = lines[pos]
            if line:
                raw, effective = classifyws(line, tabwidth)
                ntabs, nspaces = divmod(effective, tabwidth)
                lines[pos] = '\t' * ntabs + ' ' * nspaces + line[raw:]
        self._scimoz.replaceSel(''.join(lines))

    def _do_cmd_untabify(self):
        selection = self._scimoz.selText
        lines = selection.splitlines(1)
        tabwidth = self._asktabwidth()
        for pos in range(len(lines)):
            lines[pos] = lines[pos].expandtabs(tabwidth)
        self._scimoz.replaceSel(''.join(lines))

    def _do_cmd_backSmart(self):
        # if the pref is off, just do a regular delete
        if (self._scimoz.currentPos != self._scimoz.anchor or
            not self._view.prefs.getBooleanPref('editBackspaceDedents')):
            self._scimoz.deleteBack()
            return
        # If there is only whitespace to the left of us (on the current line), we delete enough characters
        #    to do a proper alignment ( this is currently done by IDLE, and works well w.r.t. tabs, spaces, etc.
        # Else we just do a regular backspace.
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        lineStart = self._scimoz.positionFromLine(lineNo)
        before = self._scimoz.getTextRange(lineStart, self._scimoz.currentPos)
        if not before or before.strip() != '':
            # we're either at the beginning of the line or
            # somewhere in the line with text to our left -- just do a backspace.
            self._scimoz.deleteBack()
            return
        # the following is taken from idle (with variable renamings)
        # here we've got only whitespace to our left.
        # Delete whitespace left, until hitting a real char or closest
        # preceding virtual tab stop.
        # Ick.  It may require *inserting* spaces if we back up over a
        # tab character!  This is written to be clear, not fast.
        tabwidth = self._scimoz.tabWidth
        indentwidth = self._scimoz.indent
        have = len(before.expandtabs(tabwidth))
        assert have > 0
        want = int((have - 1) / indentwidth) * indentwidth
        ncharsdeleted = 0
        while 1:
            before = before[:-1]
            ncharsdeleted = ncharsdeleted + 1
            have = len(before.expandtabs(tabwidth))
            if have <= want or before[-1] not in " \t":
                break
        self._scimoz.beginUndoAction()
        try:
            for i in range(ncharsdeleted):
                self._scimoz.deleteBack()
            if have < want:
                inserted = ' ' * (want - have)
                self._scimoz.replaceSel(inserted)
        finally:
            self._scimoz.endUndoAction()

    def _insertIndent(self):
        sm = self._scimoz
        currentLineNo = sm.lineFromPosition(sm.currentPos)
        lineStart = sm.positionFromLine(currentLineNo)
        toLeft = sm.getTextRange(lineStart, sm.currentPos)
        if not toLeft.strip(): # we've got nothing but whitespace to our left:
            currentIndentWidth = self._getIndentWidthForLine(currentLineNo,
                                                            sm.currentPos)
            numIndents, extras = divmod(currentIndentWidth, self._scimoz.indent)
            numIndents += 1
            newIndentWidth = numIndents * self._scimoz.indent
            newIndent = self._makeIndentFromWidth(newIndentWidth)
            sm.anchor = sm.positionFromLine(currentLineNo)
            sm.replaceSel(newIndent)
        else:
            # this isn't quite what vs.net does -- vs.net converts
            # all of the whitespace around the cursor to tabs
            # if appropriate
            startCol = sm.getColumn(sm.currentPos)
            numIndents, extras = divmod(startCol, self._scimoz.indent)
            numIndents += 1
            targetCol = numIndents * self._scimoz.indent
            if sm.useTabs:
                sm.replaceSel('\t')
            else:
                sm.replaceSel(' ' * (targetCol - startCol))

    def _insertDedent(self):
        sm = self._scimoz
        startCol = sm.getColumn(sm.currentPos)
        if self._scimoz.indent == 0:
            log.error("scimoz indent was 0, should never happen")
            return
        numIndents, extras = divmod(startCol, self._scimoz.indent)
        if numIndents and not extras:
            numIndents -= 1
        targetCol = numIndents * self._scimoz.indent
        sm.beginUndoAction()
        try:
            while (sm.getColumn(sm.currentPos) > targetCol and
                   sm.getWCharAt(sm.positionBefore(sm.currentPos)) in ' \t'):
                sm.deleteBack()
            curCol = sm.getColumn(sm.currentPos)
            if curCol < targetCol:
                d = targetCol-curCol
                sm.addText(d, d*' ')
        finally:
            sm.endUndoAction()

    def _do_cmd_dedent(self):
        sm = self._scimoz

        # Do tab autocompletion, assuming it's in progress.
        if sm.autoCActive():
            sm.autoCComplete()
            return

        if self._prefs.getBooleanPref('editTabCompletes') and \
           sm.getWCharAt(sm.positionBefore(sm.currentPos)) in self.wordchars:
            self._doCompleteWord(1)
            return

        selectionStartLine = sm.lineFromPosition(sm.selectionStart)
        endOfSelectionStartLine = sm.getLineEndPosition(selectionStartLine)
        selectionEndLine = sm.lineFromPosition(sm.selectionEnd)
        startColumn = sm.getColumn(sm.selectionStart)

        # Do we have a selection?  If no, then it's 'insert a backwards tab'
        if sm.currentPos == sm.anchor:
            if startColumn == 0:
                sm.currentPos = endOfSelectionStartLine
                self._regionShift(-1)
                sm.currentPos = sm.anchor
            else:
                self._insertDedent()
            return

        # If we have a selection, we first figure out if it's within-line or
        # either whole-line or multi-line.
        
        if (selectionStartLine != selectionEndLine or
            startColumn == 0 and sm.selectionEnd == endOfSelectionStartLine):
            return self._regionShift(-1)
        else:
            return self._insertDedent()

    wordchars = string.letters + string.digits + "_"
    
    def _do_cmd_completeWord(self):
        self._doCompleteWord(0)

    def _do_cmd_completeWordBack(self):
        self._doCompleteWord(1)

    def _doCompleteWord(self, backwards):
        # adapted from IDLE's AutoExpand code, with the additional
        # tweak of 'going backwards'
        curinsert = self._scimoz.currentPos
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineNo)
        endofLinePos = self._scimoz.getLineEndPosition(lineNo)
        curline = self._scimoz.getTextRange(startofLinePos, endofLinePos)
        if not self._completeWordState:
            words = self._getwords()
            index = 0
        else:
            words, index, insert, line = self._completeWordState
            if insert != curinsert or line != curline:
                words = self._getwords()
                index = 0
        if not words:
            return
        word = self._getprevword()
        self._scimoz.anchor = self._scimoz.currentPos - len(word)
        self._scimoz.replaceSel('')
        if backwards:
            newword = words[(index-2)  % len(words)]
            index = (index - 1) % len(words)
        else:
            newword = words[index]
            index = (index + 1) % len(words)
        self._scimoz.replaceSel(newword)
        curinsert = self._scimoz.currentPos
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineNo)
        endofLinePos = self._scimoz.getLineEndPosition(lineNo)
        curline = self._scimoz.getTextRange(startofLinePos, endofLinePos)
        
        self._completeWordState = words, index, curinsert, curline

    def _getwords(self):
        # this is where we would do magic to look through other buffers
        word = self._getprevword()
        if not word:
            return []
        before = self._scimoz.getTextRange(0, self._scimoz.currentPos)
        wbefore = re.findall(r"\b" + word + r"\w+\b", before)
        del before
        after = self._scimoz.getTextRange(self._scimoz.currentPos, self._scimoz.textLength)
        wafter = re.findall(r"\b" + word + r"\w+\b", after)
        del after
        if not wbefore and not wafter:
            return []
        words = []
        dict = {}
        # search backwards through words before
        wbefore.reverse()
        for w in wbefore:
            if dict.get(w):
                continue
            words.append(w)
            dict[w] = w
        # search onwards through words after
        for w in wafter:
            if dict.get(w):
                continue
            words.append(w)
            dict[w] = w
        words.append(word)
        return words

    def _getprevword(self):
        curinsert = self._scimoz.currentPos
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineno)
        line = self._scimoz.getTextRange(startofLinePos, curinsert)
        i = len(line)
        while i > 0 and line[i-1] in self.wordchars:
            i = i-1
        return line[i:]

    def _do_cmd_indent(self):
        sm = self._scimoz

        # Do tab autocompletion, assuming it's in progress.
        if sm.autoCActive():
            sm.autoCComplete()
            return

        if self._prefs.getBooleanPref('editTabCompletes') and \
           sm.getWCharAt(sm.currentPos-1) in self.wordchars:
            self._doCompleteWord(0)
            return

        # Do we have a selection?  If no, then it's 'insert a tab'
        if sm.currentPos == sm.anchor:
            self._insertIndent()
            return
        
        # If we have a selection, we first figure out if it's within-line or
        # either whole-line or multi-line.
        
        selectionStartLine = sm.lineFromPosition(sm.selectionStart)
        selectionEndLine = sm.lineFromPosition(sm.selectionEnd)
        startColumn = sm.getColumn(sm.selectionStart)
        endOfSelectionStartLine = sm.getLineEndPosition(selectionStartLine)
        if (selectionStartLine != selectionEndLine or
            startColumn == 0 and sm.selectionEnd == endOfSelectionStartLine):
            return self._regionShift(1)
        else:
            return self._insertIndent()

    def _regionShift(self, shift):
        """ This code shifts regions of text left or right depending on the
        sign of 'shift' (which should be +1 for a rightward shift or
        -1 for a leftward shift
        
        First we adjust the selection to consist only of whole lines.
        
        Then we look for the delta in number of spaces that should be applied
        to the first line which has a non-zero indent.
        
        Then we apply that delta to each line in the selection, ensuring that
        the indentations that result respect the use of tabs as specified
        in the widget
        """
        sm = self._scimoz
        indentlog.info("doing _regionShift by shift: %s", shift)
        # first adjust the selection to span full lines.
        anchorLine = sm.lineFromPosition(sm.anchor)
        currentPosLine = sm.lineFromPosition(sm.currentPos)
        if sm.anchor < sm.currentPos:
            anchorFirst = 1
            startLine = anchorLine
            endLine = currentPosLine
        else:
            anchorFirst = 0
            endLine = anchorLine
            startLine = currentPosLine
            # ensure that anchor starts for the sake of our computation
            # things get fixed at the end.
            sm.anchor, sm.currentPos = sm.currentPos, sm.anchor

        anchorColumn = sm.getColumn(sm.anchor)
        currentPosColumn = sm.getColumn(sm.currentPos)
        if anchorColumn != 0:
            l = sm.lineFromPosition(sm.anchor)
            sm.anchor = sm.positionFromLine(l)
        if currentPosColumn != 0:
            # adjust to end of line
            l = sm.lineFromPosition(sm.currentPos)
            sm.currentPos = sm.getLineEndPosition(l)
        # figure out what delta we need to apply to the first line
        for line in range(startLine, max(endLine, startLine+1)):
            currentIndentWidth = self._getIndentWidthForLine(line)
            if shift == 1 or currentIndentWidth:
                break
        if currentIndentWidth == 0 and shift == -1:
            return # nothing to do, nothing's indented!
        numIndents, extras = divmod(currentIndentWidth, self._scimoz.indent)
        if shift == 1:
            numIndents += 1
        elif shift == -1 and numIndents and not extras:
            numIndents -= 1
        newIndentWidth = numIndents * self._scimoz.indent
        delta = newIndentWidth - currentIndentWidth
        indentlog.info("delta = %d", delta)
        # apply that delta to each line in the region
        region = sm.getTextRange(sm.anchor, sm.currentPos)
        lines = region.splitlines(1) # keep line ends
        data = []
        for line in lines:
            count = 0
            for char in line:
                if char in ' \t':
                    count += 1
                else:
                    break
            indent = line[:count].expandtabs(self._scimoz.tabWidth)
            rest = line[count:]
            numspaces = max(len(indent) + delta, 0)
            indent = self._makeIndentFromWidth(numspaces)
            newline = indent + rest
            data.append(newline)
        region = ''.join(data)
        before = min(sm.currentPos, sm.anchor)
        sm.targetStart = sm.anchor
        sm.targetEnd = sm.currentPos
        sm.replaceTarget(len(region), region)
        if anchorFirst:
            sm.anchor = before
            sm.currentPos = before + len(region)
        else:
            sm.currentPos = before
            sm.anchor = before + len(region)

    def unexpandtabs(self, line):
        numspaces = len(line)-len(line.lstrip())
        rest = line[numspaces:]
        numtabs, numspaces = divmod(numspaces, self._scimoz.tabWidth)
        indent = '\t'*numtabs + ' '*numspaces
        return indent + rest


def classifyws(s, tabwidth):
    raw = effective = 0
    for ch in s:
        if ch == ' ':
            raw = raw + 1
            effective = effective + 1
        elif ch == '\t':
            raw = raw + 1
            effective = (effective / tabwidth + 1) * tabwidth
        else:
            break
    return raw, effective
