#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

import os, sys
import threading
import time
from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLintResult import *
from koLintResults import koLintResults

import logging
log = logging.getLogger("koLintService")
#log.setLevel(logging.DEBUG)

class RequestQueue:
    # This is a modification if Python's std Queue.Queue class:
    #   - drop maxsize related stuff
    #   - calls are always blocking
    #   - add .prepend() and .remove_uid()
    def __init__(self):
        import thread
        self._init()
        self.mutex = thread.allocate_lock()
        self.esema = thread.allocate_lock() # if acquired, then queue is empty
        self.esema.acquire()

    def put(self, item):
        """Put an item into the queue."""
        log.debug("in RequestQueue.put, acquiring mutex")
        self.mutex.acquire()
        try:
            was_empty = self._empty()
            self._append(item)
            # If we fail before here, the empty state has
            # not changed, so we can skip the release of esema
            if was_empty and self.esema.locked():
                log.debug("in RequestQueue.put, releasing esema")
                self.esema.release()
        finally:
            # Catching system level exceptions here (RecursionDepth,
            # OutOfMemory, etc) - so do as little as possible in terms
            # of Python calls.
            log.debug("in RequestQueue.put, releasing mutex")
            self.mutex.release()

    def prepend(self, item):
        """Prepend an item to the queue."""
        log.debug("in RequestQueue.prepend, acquiring mutex")
        self.mutex.acquire()
        try:
            was_empty = self._empty()
            self._prepend(item)
            # If we fail before here, the empty state has
            # not changed, so we can skip the release of esema
            if was_empty and self.esema.locked():
                log.debug("in RequestQueue.prepend, releasing esema")
                self.esema.release()
        finally:
            # Catching system level exceptions here (RecursionDepth,
            # OutOfMemory, etc) - so do as little as possible in terms
            # of Python calls.
            log.debug("in RequestQueue.prepend, releasing mutex")
            self.mutex.release()

    def get(self):
        """Remove and return an item from the queue.

        Block if necessary until an item is available.
        """
        item = -1
        while item == -1:
            log.debug("in RequestQueue.get, aquiring esema")
            self.esema.acquire()
            log.debug("in RequestQueue.get, aquiring mutex")
            self.mutex.acquire()
            release_esema = 1
            try:
                item = self._get()
                # Failure means empty state also unchanged - release_esema
                # remains true.
                release_esema = not self._empty()
            finally:
                # since we loop on failure, we don't want to unlock
                # multiple times
                if release_esema and self.esema.locked():
                    log.debug("in RequestQueue.get, releasing esema")
                    self.esema.release()
                log.debug("in RequestQueue.get, releasing mutex")
                self.mutex.release()
        return item

    def remove_uid(self, uid):
        """Remove all current requests with the given uid.

        Does not return anything.
        """
        log.debug("in RequestQueue.remove_uid, aquiring mutex")
        self.mutex.acquire()
        try:
            self._remove_uid(uid)
        finally:
            log.debug("in RequestQueue.remove_uid, releasing mutex")
            self.mutex.release()

    #---- Override these methods to implement other queue organizations
    # (e.g. stack or priority queue). These will only be called with
    # appropriate locks held.

    # Initialize the queue representation
    def _init(self):
        self.queue = []

    # Check whether the queue is empty
    def _empty(self):
        return not self.queue

    # Put a new item in the queue
    def _append(self, item):
        self.queue.append(item)
    def _prepend(self, item):
        self.queue.insert(0, item)

    # Get an item from the queue
    def _get(self):
        item = self.queue[0]
        del self.queue[0]
        return item

    # Remove all requests with the given uid.
    def _remove_uid(self, uid):
        self.queue = [request for request in self.queue if request.uid != uid]


class KoLintRequest:
    _com_interfaces_ = [components.interfaces.koILintRequest]
    _reg_desc_ = "Komodo Lint Request"
    _reg_clsid_ = "{845A872F-293F-4a82-8552-40849A92EC80}"
    _reg_contractid_ = "@activestate.com/koLintRequest;1"

    def __init__(self):
        self.document = None
        self.uid = ''
        self.linterType = ''
        self.cwd = ''
        
        self.content = None
        self.encoding = None
        self.linter = None
        self.results = None
        self.errorString = ''
        self.canceled = None
    
    def describe(self):
        return "<KoLintRequest: %s on uid %s>"\
               % (self.linterType, self.uid)


class KoLintService:
    _com_interfaces_ = [components.interfaces.koILintService]
    _reg_desc_ = "Komodo Lint Management Service"
    _reg_clsid_ = "{9FD67601-CB60-411D-A212-ED21B3D25C15}"
    _reg_contractid_ = "@activestate.com/koLintService;1"

    def __init__(self):
        log.info("KoLintService.__init__()")
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self.observerService = components.classes['@activestate.com/koObserverService;1'].\
                                createInstance(components.interfaces.nsIObserverService)
        # Let's try sync because see WARNING about PROXY_ASYNC here:
        # http://swatter.activestate.com/lxr/source//mozilla/xpcom/proxy/public/nsProxyEvent.h#61
        # We pass an XPCOM component (KoLintResult) in our notifications here.
        self.observerServiceProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService, self.observerService,
            PROXY_ALWAYS | PROXY_SYNC)

        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].getService()

        self._linterCache = {} # mapping of linterCID to koILinter instance
        self.requests = RequestQueue() # an item of None is the quit sentinel
        self.shutdown = 0
        self.manager = threading.Thread(target=self.run)
        self.manager.start()

    def terminate(self):
        log.info("KoLintService.terminate()")
        self.requests.prepend(None) # prepend the quit sentinel
        self.shutdown = 1
        self.manager.join()

    def broadcastResultsCleared(self, request):
        #XXX This should die.
        log.info("KoLintService.broadcastResultsCleared(%s)",
                 request.describe())
        self.observerService.notifyObservers(request, request.uid,
                                             "lint_clear")

    def _getLinter(self, linterCID):
        """Return a koILinter XPCOM component of the given linterCID.
        
        This method cache's linter instances. If there is no such linter
        then an exception is raised an error is set on koILastError.
        """
        if linterCID not in self._linterCache:
            try:
                linter = components.classes[linterCID].createInstance(components.interfaces.koILinter)
            except COMException, ex:
                errmsg = "Internal Error creating a linter with CID '%s': %s"\
                         % (linterCID, ex)
                self._lastErrorSvc.setLastError(100, errmsg)
                raise ServerException(nsError.NS_ERROR_UNEXPECTED, errmsg)
            else:
                self._linterCache[linterCID] = linter

        return self._linterCache[linterCID]

    def addRequest(self, request):
        """Add the given request to the queue.
        
        If there is an error (e.g. bogus linterType) an exception is raised
        and an error set on koLastErrorService.
        """
        log.info("KoLintService.addRequest(%s)", request.describe())
        
        # Fill out the request (because document access and component
        # creation must often be done in the main thread).
        request.content = request.document.buffer
        request.encoding = request.document.encoding
        if request.linterType:
            request.linter = self._getLinter(request.linterType)

        self.requests.put(request)

    def cancelRequests(self, uid):
        log.info("KoLintService.cancelRequests(uid='%s')", uid)
        self.requests.remove_uid(uid)
        #XXX This does nothing to stop the reporting of results from a
        #    possible _currently running_ lint request for this uid.
        #    This is currently handled on the JavaScript side via the
        #    koILintRequest.canceled attribute. Fine for now.

    def _getEncodingLintResults(self, content, encoding):
        """Return lint results for encoding errors in the given document.
        
            "content" is the document content as a unicode string
            "encoding" is the currently selected encoding for the document
        
        Returns a koLintResults instance.
        """
        try:
            encodedString = content.encode(encoding.python_encoding_name,
                                           "strict")
        except UnicodeError, ex:
            pass  # errors are handled after the try/except/else block
        else:
            return koLintResults() # no encoding errors
        
        # Find the specific errors by encoding with "replace" and finding
        # where those replacements were.
        escapedContent = content.replace('?', 'X')
        encodedString = escapedContent.encode(encoding.python_encoding_name,
                                              "replace")
        offset = 0
        indeces = []
        while 1:
            index = encodedString.find('?', offset)
            if index == -1:
                break
            indeces.append(index)
            offset = index + 1
        log.debug("encoding errors at indeces %s", indeces)
            
        results = koLintResults()
        lines = content.splitlines(1) # keep line terminators
        offset = 0 # the current offset in the document
        for i in range(len(lines)):
            line = lines[i]
            while indeces and indeces[0] < offset + len(line):
                index = indeces.pop(0) # this index is on this line
                r = KoLintResult()
                r.description = "This character cannot be represented with "\
                                "the current encoding: '%s'"\
                                % encoding.python_encoding_name
                r.lineStart = i+1
                r.lineEnd = i+1
                r.columnStart = index - offset + 1
                r.columnEnd = r.columnStart + 1
                log.debug("encoding error: index=%d: %d,%d-%d,%d", index,
                          r.lineStart, r.columnStart, r.lineEnd, r.columnEnd)
                r.severity = r.SEV_ERROR
                results.addResult(r)
            if not indeces:
                break
            offset += len(line)
        else:
            raise ValueError("Did not find line and column for one or "
                             "more indeces in content: %s" % indeces)

        return results

    def run(self):
        """Process lint requests serially until told to stop.
        
        Before the requested linter is run on a document it is first checked
        for encoding problems (i.e. encoding is not sufficient for current
        content).
        """
        log.info("manager thread: start")
        while 1:
            try:
                # wait for next request
                request = self.requests.get()
                
                # quit if request is the quit sentinel
                if request is None:
                    log.info("manager thread: quit sentinel")
                    break
    
                # process the request
                log.info("manager thread: process request: %r", request)
                self._lastErrorSvc.setLastError(0, None) # clear last error
                try:
                    # Look for encoding errors first.
                    results = self._getEncodingLintResults(request.content,
                                                           request.encoding)
    
                    # If there were no encoding errors, try the reqeusted linter.
                    if not results.getNumResults() and request.linter:
                        #XXX This is where context-sensitive linting args should
                        #    be passed in, but linters don't support this yet.
                        results = request.linter.lint(request)
    
                    request.results = results
                except (ServerException, COMException), ex:
                    errno, errmsg = self._lastErrorSvc.getLastError()
                    if errno == 0 and not errmsg:
                        err = "unexpected error checking '%s' with '%s' linter"\
                              % (request.document.baseName, request.linterType)
                        log.exception(err)
                    else:
                        err = "Syntax checking error: "+errmsg
                    request.errorString = err
                except:
                    # Any exceptions that are not ServerException or COMException
                    # are unexpected internal errors.
                    err = "unexpected internal error checking '%s' with '%s' linter"\
                          % (request.document.baseName, request.linterType)
                    log.exception(err)
                    request.errorString = err
                else:
                    log.info("manager thread: lint results for uid %s: %r",
                             request.uid, results)

                if self.shutdown:
                    break

                # notify of request completion
                self.observerServiceProxy.notifyObservers(request,
                                                          request.uid,
                                                          "lint_results")
            except:
                # something bad happened, but dont let this thread die
                pass
            
        log.info("manager thread: end")



if __name__ == "__main__":
    logging.basicConfig()
    import pprint
    class TestRequest:
        def __init__(self, uid):
            self.uid = uid
        def __repr__(self):
            return "<TestRequest: uid=%s>" % self.uid
    q = RequestQueue()

    if 0:
        q.put(TestRequest("id_1"))
        q.remove_uid("id_1")
        print "item:"
        sys.stdout.flush()
        print q.get()
    
    if 1:    
        q.put(TestRequest("id_1"))
        q.put(TestRequest("id_2"))
        pprint.pprint(q.queue)
        print "item: ", q.get()
        q.put(TestRequest("id_3"))
        q.put(TestRequest("id_4"))
        q.put(TestRequest("id_3"))
        q.prepend(None)
        pprint.pprint(q.queue)
        q.remove_uid("id_3")
        pprint.pprint(q.queue)
        q.remove_uid("id_3")
        sys.stdout.flush()
        pprint.pprint(q.queue)
        q.remove_uid("id_4")
        pprint.pprint(q.queue)
        print "item: ", q.get()
        print "item: ", q.get()
        pprint.pprint(q.queue)

