# Copyright (c) 2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# The debug engine for Requester based interfaces

from xpcom import components, _xpcom, ServerException, COMException, nsError
#from xpcom.server.enumerator import SimpleEnumerator
import threading, thread
import os
import sys
import kdb
import dbgutils # For the engine base class.
import socket
import xmlrpclib
import urlparse
import urllib
import struct
import time
import string
import types

RESUME_STOP = components.interfaces.koIDebugger.RESUME_STOP
RESUME_STEP_IN = components.interfaces.koIDebugger.RESUME_STEP_IN
RESUME_STEP_OVER = components.interfaces.koIDebugger.RESUME_STEP_OVER
RESUME_STEP_OUT = components.interfaces.koIDebugger.RESUME_STEP_OUT
RESUME_GO = components.interfaces.koIDebugger.RESUME_GO

DBG_STATE_STOPPED = components.interfaces.koIDebugger.DBG_STATE_STOPPED
DBG_STATE_BREAK = components.interfaces.koIDebugger.DBG_STATE_BREAK
DBG_STATE_RUNNING = components.interfaces.koIDebugger.DBG_STATE_RUNNING
DBG_STATE_STARTING = components.interfaces.koIDebugger.DBG_STATE_STARTING
DBG_STATE_STOPPING = components.interfaces.koIDebugger.DBG_STATE_STOPPING

import logging
log = logging.getLogger('debugger.requester')

# created in the engines - xxx - bogus!

class koRequesterBreakpoint(dbgutils.koBreakpointBase):
    def __init__(self, language, engine, where, line):
        self.number = -1 # The number the remote process knows us by
        self.language = language
        dbgutils.koBreakpointBase.__init__(self, engine, where, line)

    def get_language(self):
        return self.language
    
    def _addBreakpoint(self):
        return self.engine.requester.addBreakpoint(file=str(self.remoteURI), lineno=self.lineno, temporary = xmlrpclib.False, cond = "");
    
    def _updatePending(self):
        if not self.pending:
            return

        requester = self.engine.requester
        # XXX this is somewhat of a hack to work around lifetime issues.
        try:
            requester.hello()
        except:
            log.error("could not talk to requestor!")
            return

        self.pending = 0
        if self.state == components.interfaces.koIBreakpoint.BREAKPOINT_DELETED:
            if self.number != -1:
                requester.removeBreakpoint(breakpointNumber=self.number)
                self.number = -1
            return
        # Either addition or modification - ensure we have a number
        
        if self.number == -1:
            self.number = self._addBreakpoint()
            if self.number == -1:
                self.state = components.interfaces.koIBreakpoint.BREAKPOINT_DISABLED
                # Eeek - no breakpoint at that location - just set as disabled, but not pending.
                return

        # OK - valid breakpoint with some state -set it.
        enabled = (self.state == components.interfaces.koIBreakpoint.BREAKPOINT_ENABLED)
        # XXX This is commented out until we decide if enableBreakpoint will superceed setBreakpointAttribute
        # Either way, the call is not required yet because Komodo does not (yet) disable breakpoints
        # and they are enabled by default
        # requester.enableBreakpoint(breakpointNumber=self.number, enable=xmlrpc.Boolean(enabled))



class koRequesterDebugProperty:
    _com_interfaces_ = components.interfaces.koIDebugProperty
    # If this object needs to be registered, the following 2 are also needed.
    # _reg_clsid_ = {a new clsid generated for this object}
    # _reg_contractid_ = "The.Object.Name"
    def __init__(self, language, requester, localArgs, remoteProperty, context = None):
        self.language = language
        self.requester = requester
        self.frame_index = localArgs['frame_index']
        self.radix = localArgs['radix']
        self.short_name = remoteProperty['name']['name']
        self.expr = remoteProperty['name']['fullName']
        self.value = remoteProperty['value']
        self.type = remoteProperty['type']
        self.has_children = remoteProperty['hasChildren']
        self.context = context

    def get_language(self):
        return self.language
    def get_name( self ):
        # Result: wstring
        return self.short_name
    def get_fullName( self ):
        # Result: wstring
        return self.expr
    def get_type( self ):
        # Result: wstring
        return self.type
    def get_value( self ):
        # Result: wstring
        # XXX - todo - use self.radix
        # note: binary data kills komodo, repr prevents that.
        return repr(self.value)[1:-1] # ignore quotes
    def get_hasChildren( self ):
        # Result: boolean
        return self.has_children
    def SetValueAsString( self, value):
        # Result: void - None
        # In: value: wstring
        # Out: err_msg: wstring

        # XXX do not allow unicode
        # once xml-rpc can handle unicode, we can decide whether
        # to force languages that do not handle unicode to do the conversion
        if isinstance(value, types.UnicodeType): 
            try:
                value = str(value)
            except UnicodeError:
                return "%s can only evaluate expressions encoded in ASCII" % self.get_language( )

        result = self.requester.setDebugProperty(self.context,
                                                 frame_index = self.frame_index,
                                                 lhs = self.expr,
                                                 rhs = value,
                                                 radix = 10)
        if result:
            return result
        else:
            # Update our value.
            # XXX This is wrong. Should update all state info. And we aren't passing the correct context here! 
            propertyInfo = self.requester.getDebugPropertyInfo(frame_index=self.frame_index, expr=self.expr, radix=self.radix)
            
            self.value = propertyInfo.value
            self.type = propertyInfo.type
            self.has_children = propertyInfo.has_children

        return None # No error.

    def enumChildren(self):
        # note that expr should be Fully Qualified Name 
        # (suitable for 'eval'ing)
        children = self.requester.getChildrenInfos(self.context, frame_index=self.frame_index,parent_expr=self.expr, radix=self.radix)
        #return SimpleEnumerator(children)
        return children

class koRequesterDebugStackFrame:
    _com_interfaces_ = components.interfaces.koIDebugStackFrame

    def __init__(self, language, requester, constructUri, frame_index, localArgs, remoteFrame):
        self.language = language
        self.requester = requester
        self.frame_index = frame_index
        self.baseURI = remoteFrame['file']
        self.uri = constructUri( self.baseURI )
        self.where = remoteFrame['where']
        self.lineno = remoteFrame['lineno']

        self.inputBaseURI = ""
        self.inputURI = ""
        self.inputLineno = 0
        self.inputWhere = ""
        if remoteFrame.has_key('data'):        
            self.inputBaseURI = remoteFrame['data']['file']
            if self.inputBaseURI:
                self.inputURI = constructUri( self.inputBaseURI )
            self.inputLineno = remoteFrame['data']['lineno']
            self.inputWhere = remoteFrame['data']['where']        

    def get_inputURI(self):
        return self.inputURI
    
    def get_inputLineNumber(self):
        return self.inputLineno

    def get_language(self):
        return self.language
    
    def get_uri(self):
        return self.uri

    def get_filename(self):
        # Result: wstring
        # this is possibly true, but slowing down debugging =(
        # log.error("Use koPythonDebugStackFrame.uri rather than .filename")
        # XXX Remove this, no one should use it. For now it is in because there is a check for
        # it to determine if the file is really a Perl "eval"
        return ''

    def get_lineNumber(self):
        # Result: uint32
        return self.lineno

    def get_description(self):
        full_where = ''
        if self.where:
            full_where = ", in %s" % (self.where,)

        return "%s, line %d%s" % (string.split( self.baseURI, '/' )[-1], self.lineno, full_where)

    def getContextNames(self):
        return self.requester.getContextNamesAsString( frame_index = self.frame_index).split()

    def enumContext(self, context):
        # Result: nsISimpleEnumerator
        # In: param0: string

        #return SimpleEnumerator(self.requester.getContext(context=context, frame_index=self.frame_index, radix=10))
        return self.requester.getContext(context=context, frame_index=self.frame_index, radix=10)

    def getExpression(self, expr, radix):
        return self.requester.getDebugPropertyInfo(frame_index=self.frame_index, expr=expr, radix=radix)

# Handles explicit redirection initiated by debugger.
class SocketIOHandler:
    def __init__(self, stdin, stderr, stdout):
        # spawn off a thread for each of the files
        if stdin is not None:
            self.stdin_port, self.stdin_thread = self._SetupIOThread(self._SocketInputThread, stdin);
        else:
            self.stdin_port = self.stdin_thread = None
        if stdout is not None:
            self.stdout_port, self.stdout_thread = self._SetupIOThread(self._SocketOutputThread, stdout)
        else:
            self.stdout_port = self.stdout_thread = None
        if stderr is not None:
            self.stderr_port, self.stderr_thread = self._SetupIOThread(self._SocketOutputThread, stderr);
        else:
            self.stderr_port = self.stderr_thread = None

    def _SetupIOThread(self, handler, file):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET,
                        socket.SO_LINGER,
                        struct.pack("ii", 1, 30))
        sock.bind(('', 0))
        #print "listen"
        sock.listen(1)
        port = sock.getsockname()[1]
        thread = threading.Thread(target=handler, 
                                            args=(sock,file))
        thread.start()
        return port, thread

    def _SocketOutputThread(self, sock, file):
        #print "Waiting for connection"
        conn,addr = sock.accept()
        while 1:
            try:
                data = conn.recv(1024)
            except socket.error:
                break
            if not data: 
                log.debug("No data in socket output thread - _SocketOutputThread terminating")
                break
            file.write(data)

    def _SocketInputThread(self, sock, file):
        conn,addr = sock.accept()
        while 1:
            data = file.read(1024)
            if not data:
                break
            conn.send(data)
   

class RequesterProxy:
    def __init__(self, language, constructUri, cookie, url ):
        self.language = language
        self.constructUri = constructUri
        self.cookie = cookie
        self.requester = self._createRequester( url )

    def _createRequester( self, url ):            
        return xmlrpclib.Server( url )
        
    def _createStackFrame( self, frame_index, args, remoteFrame ):
        return koRequesterDebugStackFrame( self.language, self, self.constructUri, frame_index, args, remoteFrame )

    def _createDebugProperty( self, args, remoteInfo, context = None ):
        return koRequesterDebugProperty( self.language, self, args, remoteInfo, context )
    
    def _doCall(self, function, args ):
        #import time
        
        args.update( {'cookie' : self.cookie } )
        #print ('Calling %s with args:' % repr(function) ), repr(args)
        #start = time.clock( )
        result = apply( getattr( self.requester.Requester, function ), (args,) )
        #stop = time.clock( )
        #print 'Returned:', repr(result)
        #print 'Call took %s milliseconds' % ((stop - start) * 1000)
        return result
    
    def hello(self, **args):
        return self._doCall('hello', args)['status'] 

    def getNumFrames(self, **args):
        return self._doCall('getNumFrames', args)['numFrames']

    def getFrames(self, **args):
        remoteFrames = self._doCall('getFrames', args)['frames']
        localFrames = []
        
        for frame_index in range(len(remoteFrames)):
            localFrames.append( self._createStackFrame( frame_index, args, remoteFrames[frame_index] ) )

        return localFrames

    def getContextNamesAsString(self, **args):
        return self._doCall('getContextNamesAsString',args)['names']

    def _remoteToLocalPropertyList( self, args, propertyInfoList, context ):
        propertyList = []
        
        for info in propertyInfoList:
            propertyList.append( self._createDebugProperty( args, info, context ) )

        return propertyList
       
    def getContext(self, context, **args):
        args.update( { 'context' : context } )
        propertyInfoList = self._doCall('getContext',args)['propertyInfo']
        
        return self._remoteToLocalPropertyList( args, propertyInfoList, context )    

    def resumeExecution(self, **args):
        return self._doCall('resumeExecution',args)

    def getDebugPropertyInfo(self, **args):
        propertyInfo = self._doCall('getDebugPropertyInfo',args)['propertyInfo']
        return self._createDebugProperty( args, propertyInfo )

    def getChildrenInfos( self, context = None, **args ):
        if context is not None:
            args.update( { 'context' : context } )
        else:
            args.update( { 'context' : "" } )
        propertyInfoList = self._doCall('getChildrenInfos',args)['propertyInfo']
        return self._remoteToLocalPropertyList( args, propertyInfoList, context )

    def addBreakpoint( self, **args ):
        return self._doCall( 'addBreakpoint', args )['breakpointNumber']

    def removeBreakpoint( self, **args ):
        return self._doCall( 'removeBreakpoint', args )

    def setDebugProperty( self, context = None, **args ):
        if context is not None:
            args.update( { 'context' : context } )
        else:
            args.update( { 'context' : "" } )
        return self._doCall( 'setDebugProperty', args )['setResultString']

    def getSourceCode( self, **args ):
        return self._doCall( 'getSourceCode', args )['source']
    
    def redirectStdin(self, **args):
        return self._doCall( 'redirectStdin', args )

    def redirectStdout(self, **args):
        return self._doCall( 'redirectStdout', args )

    def redirectStderr(self, **args):
        return self._doCall( 'redirectStderr', args )
                    
def constructURL( originalUrl, prefix, exceptForProtocols = [] ):
    originalProtocol, junk = urllib.splittype( originalUrl )

    if originalProtocol in exceptForProtocols:
        return originalUrl
    else:
        return prefix + urllib.quote( originalUrl )
    
class koRequesterDebugEngine(dbgutils.koDebugEngineBase):
    baseURI = None
    _com_interfaces_ = components.interfaces.koIDebugEngine
    
    def __init__(self, language):
        self.language = language
        self.stopping = 0
        dbgutils.koDebugEngineBase.__init__(self)

    def get_language(self):
        return self.language
    
    def startDebugging(self, stepType, preferences, stdin, stdout, stderr):
        breakpoints=self.breakpoints
        self.breakpoints=[]
        # some lock needs to be cleared
        self.__init__()

        self.stopping = 0
        if stdin is not None or stdout is not None or stderr is not None:
            # remote debugging
            self.socketIOHandler = SocketIOHandler(stdin, stdout, stderr)
            exclude_protocols = ""
            # don't restore breakpoints for remote debugging
        else:
            exclude_protocols = "file"
            self.socketIOHandler  = None
            # keep old breakpoints
            self.breakpoints=breakpoints
        self._refreshBreakpoints()

        handler = components.classes[
                        "@mozilla.org/network/protocol;1?name=kodebugger"].\
                  getService(components.interfaces.koIDebuggerProtocolHandler)
        handler.registerEngine(self, self.get_language())

        self.uriConstructor = lambda originalUrl, newProtocol = self.baseURI.spec , exceptForProtocols = [exclude_protocols]: \
                                constructURL( originalUrl, newProtocol, exceptForProtocols )

        log.debug("Starting debug thread") # don't take out this warning
                                            # it initializes the log.warn
                                            # proxy
        threading.Thread(target=self._startDebugging, args=(stepType, preferences)).start()
        return 1

    def _startDebugging(self, stepType, preferences):
        try:
            self._doDebuggerInteraction(stepType, preferences)
        finally:
            self._cleanupAfterRun()

    def _cleanupAfterRun(self):                
        self._set_state(DBG_STATE_STOPPED, components.interfaces.koIDebugger.STOP_REASON_STOPPED)

    def _doDebuggerInteraction(self, stepType, preferences):
        self._set_state(DBG_STATE_STARTING, 0)
        # Give the process time to start
        if preferences.hasPref("debugger.timeout"):
            timeout = preferences.getLongPref("debugger.timeout")
        else:
            timeout = 20
        timeout /= 2 # We sleep for 2 secs per attempt

        # cookie must not be unicode!        
        cookie = str(preferences.getStringPref("debugger.listener-reponse.cookie"))
        port = preferences.getLongPref("debugger.listener-reponse.port")
        machine = str(preferences.getStringPref("debugger.listener-reponse.host"))
        url="http://%s:%s"% (machine,port)

        for i in range(timeout):
            log.debug("attempting connect...%s"%i)
            if self.stopping or self.state == DBG_STATE_STOPPED:
                # Something bad happened :-(
                log.error("Something bad happened")
                self.stopping = 1
                break
            try:
                self.requester = self._createRequesterProxy(cookie, url)
                # We block here until we can issue _some_ command.
                log.debug("Attempting hello")
                self.requester.hello()
                if self.socketIOHandler is not None:
                    log.debug("Redirecting remote debug ports")
                    port = self.socketIOHandler.stdout_port
                    if port: self.requester.redirectStdout(port=port)
                    
                    #XXX this is necessary until we have some kind of options
                    # handling with this protocol.  If we are remote debugging a web app,
                    # we cannot expect stdin to be redirected.
                    # so, in essence, we will attempt to continue with the debug session
                    # if either stdin or stderr redirection fails
                    try:
                        port = self.socketIOHandler.stdin_port
                        if port: self.requester.redirectStdin(port=port)
                    except xmlrpclib.Error, e:
                        log.error("xmlrpclib.Error redirecting stdin:" + repr(e))
                        pass
                    try:
                        port = self.socketIOHandler.stderr_port
                        if port: self.requester.redirectStderr(port=port)
                    except xmlrpclib.Error, e:
                        log.error("xmlrpclib.Error redirecting stdin:" + repr(e))
                        pass
                break
            except socket.error:
                log.debug("socket.error (will retry).")
                time.sleep(2)
            except xmlrpclib.Error, e:
                log.error("xmlrpclib.Error (will retry):" + repr(e))
                time.sleep(2)            
            except Exception, e:
                log.error("Launch failure:" + repr(e))
                break                
        else:
            log.error("Gave up waiting for debug process to start")
            self.stopping = 1

        # We definitely have a remote connection, but depending on how
        # we connected and how the remote debugger sprang into life,
        # we may or may not already be debugging. If we are not, the
        # remote debugger is expecting a resume command to tell it how
        # to _start_.  If it is already debugging, then the resume
        # command operates as-per normal.
        # Either way, the remote debugger is definitely running a
        # command loop, and needs a "resume" command to get crankin'
        self._processPendingBreakpoints()
        if not self.stopping:
            # We can not consider ourself running
            self._set_state(DBG_STATE_RUNNING, 0)
            is_debugging = self.requester.getNumFrames() != 0
            if is_debugging:
                pass # Already ready for our first interaction
            else:
                try:
                    # Tell the debugger how we should begin...
                    self.requester.resumeExecution(request=stepType)
                except xmlrpclib.Error:
                    self.stopping = 1
                except socket.error:                
                    self.stopping = 1
                except (IOError,Exception), e:
                    self.stopping = 1
                    log.debug("Failure during resumeExecution: "+ repr(e))
                # And again, block here until we get some response back.
                # This means we are ready for our first user interaction
                # if we are in RUN state, execution will finish before we can
                # say hello.  catch that and stop
                try:
                    if not self.stopping:
                        self.requester.hello()            
                except xmlrpclib.Error:
                    self.stopping = 1
                except socket.error:                
                    self.stopping = 1
                except (IOError,Exception), e:
                    self.stopping = 1
                    log.debug("Failure during hello: "+ repr(e))

        while not self.stopping:
#            log.debug("In debugger loop!")
            try:
                resume = self._runCommandLoop(components.interfaces.koIDebugger.BREAK_REASON_BREAK)
                self.requester.resumeExecution(request=resume)
                # And again, block here until we get some response back.
            except xmlrpclib.Error:
                self.stopping = 1
            except socket.error:                
                self.stopping = 1
            except (IOError,Exception), e:
                self.stopping = 1
                log.debug("Failure during resumeExecution: " + repr(e))

            try:
                if not self.stopping:
                    self.requester.hello()            
            except xmlrpclib.Error:
                self.stopping = 1
            except socket.error:                
                self.stopping = 1
            except (IOError,Exception), e:
                self.stopping = 1
                log.debug("Failure during hello: "+ repr(e))

    def get_state(self):
        return self.state

    def _enumStackFrames( self ):
        # Result: nsISimpleEnumerator
        if self.stack_frame_list is None:
            self.stack_frame_list = []
            self.stack_frame_list = self.requester.getFrames()
            #log.warn("frame_infos = " + str(frame_infos))

        #return SimpleEnumerator(self.stack_frame_list)
        return self.stack_frame_list

    def _getSourceCode(self, filename):
        return self.requester.getSourceCode(filename=str(filename))

    def _stop( self ):
        # Result: void - None
        if not self.stopping:
            self.stopping = 1
            if self.get_state() != DBG_STATE_STOPPED:
                self._set_state(DBG_STATE_STOPPING, 0)

    def breakNow( self ):
        # Result: void - None
        pass

    def addBreakpoint( self, where, lineNumber ):
        ret = self._createBreakpoint(where, lineNumber)
        # We must add the BP to the list before updating it!
        self.breakpoints.append(ret)
        ret._tryUpdatePending()
        return ret

    def getContextNames(self):
        return self.requester.getContextNamesAsString(frame_index=self.frame_index).split()

    def _createBreakpoint( self, where, line ):
        return koRequesterBreakpoint( self.language, self, where, line )

    def _createRequesterProxy( self, cookie, url ):
        return RequesterProxy( self.language, self.uriConstructor, cookie, url )
